#!/usr/bin/env python3

from pathlib import Path
import numpy as np
import json
import ebesir as sir
import tools


if __name__ == '__main__':
    # estimate parameters for both countries
    # and apply parameters from country 2 as a scenario to country 1
    # country 1 fixed to Germany for the moment due to econ. model
    filename_cfg = tools.cmd_line_name()
    with open(filename_cfg) as f:
        cfg = json.load(f)
    start_date = tools.str_to_date(cfg['start_date'])
    end_date = tools.str_to_date(cfg['end_date'])
    cfg1 = cfg['country1']
    name = cfg1['name']
    npi_dates = tools.list_of_str_to_dates(cfg1['npi_dates'])
    cfg2 = cfg['country2']
    name2 = cfg2['name']
    npi_dates2 = tools.list_of_str_to_dates(cfg2['npi_dates'])
    npi_dates_transfer = tools.list_of_str_to_dates(cfg2['npi_dates_transfer'])
    p1, p2 = cfg2['transfer_indices']

    fixed_recovery = cfg['fixed_recovery']
    transfer = cfg['transfer']
    transfer_risk_aversion = cfg['transfer_risk_aversion']
    impl_duration = cfg['impl_duration']

    if transfer_risk_aversion:
        output_path = Path(f'{name2}-transfer_risk')
    else:
        output_path = Path(name2)
    output_path.mkdir(parents=True, exist_ok=True)

    param_names = [f'beta{i}' for i in range(1, len(npi_dates)+2)] + ['gamma', 'v']

    # Country 1 ###############################################################
    print(f'Parameter estimation for the focal nation {name}')
    # initial guesses, not sensitive at all, therefore hardcoded
    param0_besir = [
        0.5, # beta1
        0.5, # beta2
        0.2, # beta3
        0.1, # beta4
        0.1, # beta5
        0.2, # beta6
        0.2, # beta7
        1/7, # gamma
        0.00001, # v
    ]

    dates, I_in, S, I, I_obs, param, tau, beta, gamma, v, N, cost = sir.optimize(
        name,
        start_date,
        end_date,
        param0_besir,
        impl_duration,
        npi_dates,
        fixed_recovery,
    )

    tools.write_param(param, param_names, output_path / f'{name}-param.csv')
    tools.write_data(
        name,
        dates,
        npi_dates,
        start_date,
        end_date,
        N,
        impl_duration,
        I_in,
        S,
        I,
        I_obs,
        tau,
        beta,
        gamma,
        v,
        output_path / f'{name}-results.csv',
    )
    tools.write_npi_dates(npi_dates, output_path / f'{name}-npi_dates.csv')

    # Country 2 ###############################################################
    print(f'Parameter estimation for the reference nation {name2}')
    # initial guesses, not sensitive at all, therefore hardcoded
    param0_besir2 = [
        0.3, # beta1
        0.5, # beta2
        0.1, # beta3
        0.05, # beta4
        0.05, # beta5
        0.05, # beta6
        0.05, # beta7
        1/7, # gamma
        0.00001, # v
    ]

    dates2, I_in2, S2, I2, I_obs2, param2, tau2, beta2, gamma2, v2, N2, cost2 = sir.optimize(
        name2,
        start_date,
        end_date,
        param0_besir2,
        impl_duration,
        npi_dates2,
        fixed_recovery,
    )

    tools.write_param(param2, param_names, output_path / f'{name2}-param.csv')
    tools.write_data(
        name2,
        dates2,
        npi_dates2,
        start_date,
        end_date,
        N2,
        impl_duration,
        I_in2,
        S2,
        I2,
        I_obs2,
        tau2,
        beta2,
        gamma2,
        v2,
        output_path / f'{name2}-results.csv',
    )
    tools.write_npi_dates(npi_dates2, output_path / f'{name2}-npi_dates.csv')

    # Country 1 with Scenario Country 2 #######################################
    print(f"Parameter estimation for {name}'s {name2} scenario")
    param_alt = param[:]
    if transfer == 'copy':
        param_alt[p1:p2] = param2[p1:p2]
    elif transfer == 'relative':
        for i in range(p1, p2):
            param_alt[i] = param_alt[i-1] * param2[i] / param2[i-1]
    else:
        raise ValueError(f'Unknown transfer type {transfer}')

    I0 = max(I_obs[0], 1)

    t, npi_t = tools.dates_2_days(dates, npi_dates_transfer)
    if transfer_risk_aversion:
        _, risk_t = tools.dates_2_days(dates, [npi_dates_transfer[p1-1]])
        param_alt = np.append(param_alt, param2[-1])
    else:
        _, risk_t = tools.dates_2_days(dates, [])
        param_alt = np.append(param_alt, param[-1])

    I_in_alt, S_alt, I_alt, tau_alt, beta_alt, v_alt = sir.solve_besir_model_risk_transfer(
        sir.besir_risk_transfer,
        t,
        I0,
        N,
        npi_t,
        risk_t,
        param_alt,
        impl_duration,
    )
    gamma_alt = param_alt[-2]

    tools.write_data(
        name,
        dates,
        npi_dates_transfer,
        start_date,
        end_date,
        N,
        impl_duration,
        I_in_alt,
        S_alt,
        I_alt,
        I_obs,
        tau_alt,
        beta_alt,
        gamma_alt,
        v_alt,
        output_path / f'{name}-{name2} Scenario-results.csv',
        (p1, p2),
    )
    tools.write_npi_dates(
        npi_dates_transfer,
        output_path/f'{name}-{name2} Scenario-npi_dates.csv',
        (p1, p2),
    )
