#!/usr/bin/env python3

from sys import float_info
from datetime import date
import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import least_squares
import pycocap as pyc
import tools


def behavior_beta(I, tau, v, epsilon=0.7):
    """Calculate behavioral contact rate according to Quaas et al.
    """
    return tau / (1. + (1-epsilon)/epsilon * v * I)**(1. / (1.-epsilon))

def besir(t, y, N, contact_pw, p):
    """The behavioral SIR equations"""
    S, I = y

    beta = behavior_beta(I, contact_pw[int(t)], p[-1])

    S_to_I = beta / N * I * S
    dSdt = -S_to_I
    dIdt = S_to_I - p[-2] * I
    return dSdt, dIdt

def besir_risk_transfer(t, y, N, contact_pw, v_pw, gamma):
    """The behavioral SIR equations"""
    S, I = y

    beta = behavior_beta(I, contact_pw[int(t)], v_pw[int(t)])

    S_to_I = beta / N * I * S
    dSdt = -S_to_I
    dIdt = S_to_I - gamma * I
    return dSdt, dIdt

def _pp_model(sol, contact_pw, risk, N):
    S, I = sol.y[0,:], sol.y[1,:]
    beta_pw = behavior_beta(I, contact_pw, risk)
    I_in = beta_pw / N * S * I
    return I_in, S, I, contact_pw, beta_pw

def solve_besir_model(model, t, I0, N, npi_t, param, impl_duration):
    """Set up everything for solving the model and return vars of interest"""
    S0 = N - I0
    y0 = (S0, I0)

    contact_pw = tools.param_piecewise(t, param[:-2], npi_t, impl_duration)

    sol = solve_ivp(
        lambda t, y: model(t, y, N, contact_pw, param),
        (0.0, t[-1]),
        y0,
        t_eval=t,
        max_step = 1.0,
        method='RK45',
    )
    return _pp_model(sol, contact_pw, param[-1], N)

def solve_besir_model_risk_transfer(model, t, I0, N, npi_t, risk_t, param, impl_duration):
    """Set up everything for solving the model and return vars of interest"""
    S0 = N - I0
    y0 = (S0, I0)

    contact_pw = tools.param_piecewise(t, param[:-3], npi_t, impl_duration)
    risk_pw = tools.param_piecewise(t, param[-2:], risk_t, impl_duration)

    sol = solve_ivp(
        lambda t, y: model(t, y, N, contact_pw, risk_pw, param[-3]),
        (0.0, t[-1]),
        y0,
        t_eval=t,
        max_step = 1.0,
        method='RK45',
    )

    return *_pp_model(sol, contact_pw, risk_pw, N), risk_pw

def get_data(name, start_date, end_date):
    """Convenience wrapper for pycocap"""
    df, N = pyc.cases(
        name,
        start_date,
        end_date,
        params={'source_name': 'JHU'},
        fill_value='interpolate',
        return_population=True,
    )
    dates = np.array(df['date'], dtype='datetime64[D]')
    I_obs = np.array(df['cases'].fillna(0.0))
    # exclude trivial solutions
    I0 = max(I_obs[0], 1)
    return dates, I_obs, N, I0

def optimize(
    name,
    start_date,
    end_date,
    param0,
    impl_duration,
    npi_dates,
    fixed_recovery=None,
):
    """Set up everything and estimate optimal parameters"""

    dates, I_obs, N, I0 = get_data(name, start_date, end_date)
    t, npi_t = tools.dates_2_days(dates, npi_dates)

    # all parameters are positive
    bounds = (
        [0.0] * len(param0),
        [np.inf] * len(param0),
    )
    # fix the recovery rate to literature value
    if fixed_recovery is not None:
        bounds[0][-2] = fixed_recovery
        bounds[1][-2] = fixed_recovery + float_info.epsilon
    res = least_squares(
            lambda param, I_obs:
                solve_besir_model(
                    besir,
                    t,
                    I0,
                    N,
                    npi_t,
                    param,
                    impl_duration,
                )[0] - I_obs,
        param0,
        jac='2-point',
        bounds=bounds,
        args=(I_obs,),
    )
    param, cost = res.x, res.cost
    I_in, S, I, tau_pw, beta_pw = solve_besir_model(besir, t, I0, N, npi_t, res.x, impl_duration)
    gamma = param[-2]
    v = param[-1]
    return dates, I_in, S, I, I_obs, param, tau_pw, beta_pw, gamma, v, N, cost


if __name__ == '__main__':
    pass
