#!/usr/bin/env python3

from pathlib import Path
from datetime import timedelta
import json
import numpy as np
from scipy.stats import linregress
import pandas as pd
import matplotlib.pyplot as pt
from dateutil.relativedelta import relativedelta
import tools


# all parameters hardcoded, as this is a very specific model
SECTIONS = np.array([
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'OPQ',
    'RSTU',
])

K_B = np.array([
    100,
    100,
    81.2455975,
    100,
    100,
    91.5271575,
    76.4904875,
    70.0914275,
    39.8281225,
    74.717525,
    100,
    87.2584175,
    80.9673825,
    64.8319325,
    100,
    59,
]) / 100

K_B2 = np.array([
    100,
    100,
    75.7419425,
    100,
    100,
    90.2012625,
    90.0721825,
    74.1400175,
    45.2386775,
    80.908205,
    100,
    88.7039275,
    86.0225775,
    68.8631125,
    100,
    60.7,
]) / 100

R_B = np.array([
    1,
    1,
    9.2,
    1,
    1,
    8.4,
    8.4,
    9.1,
    10.5,
    8.2,
    1,
    9,
    8.9,
    8.9,
    1,
    8.8,
])

ALPHA_WEIGHTS = np.array((
    0.923236542920576,
    0.145842629574777,
    22.8,
    1.7,
    1.05025856098289,
    4.71394723658672,
    10,
    4.4,
    1.6,
    4.58528269242878,
    4.03277797701011,
    10.5591843215409,
    6.39175342398252,
    5.1,
    18.2,
    3.8,
))

EXOGENOUS_SECTIONS = ['G', 'H', 'I', 'RSTU']
EXOGENOUS_SECTIONS_INDEX = [6, 7, 8, 15]

EXCLUDE_SECTIONS = ['endog_A', 'endog_B', 'endog_D', 'endog_E', 'endog_K', 'endog_OPQ']
EXCLUDE_SECTIONS_INDEX = [0, 1, 3, 4, 10, 14]

KEY_DATES = ('2020-03-19', '2020-04-20')
NPI_IMPLEMENTATION_DURATION = 21
DURATIONS_S_LEONTIEF = 30
SIMULATION_HORIZON = 1096


def cross_sectional_parameters(Rt2, r_d_m, r_d_r):
    no_scenarios = len(Rt2)
    no_sections = len(SECTIONS)
    r_d_s = np.array((Rt2)) - r_d_r[1]
    slope = linregress(r_d_r, r_d_m).slope
    r_d_pred = r_d_s * slope
    k_slopes = (K_B2 - K_B) / (r_d_r[0] - r_d_r[1])

    k_s = np.empty((no_sections, no_scenarios))
    r_s = np.empty_like(k_s)
    delta = np.empty_like(k_s)

    for section_id in range(no_sections):
        for scenario_id in range(no_scenarios): 
            k_s[section_id, scenario_id] = max(0, min(1, K_B[section_id] + r_d_s[scenario_id] * k_slopes[section_id]))
            r_s[section_id, scenario_id] = max(1, R_B[section_id] + r_d_pred[scenario_id]) # adjusted expected duration until pre-shutdown activity at time of survey
            
            if section_id in EXOGENOUS_SECTIONS_INDEX:
                delta[section_id, scenario_id] = k_s[section_id, scenario_id] - K_B[section_id] # subtracting pre-opening economic activity level, also for endogenous sections
            else:
                delta[section_id, scenario_id] = 0

    return k_s, r_s, delta

def compute_delta_exog(Rt2, delta, leontief_mat):
    no_scenarios = len(Rt2)
    no_sections = len(SECTIONS)

    delta_mat = np.empty((no_scenarios, no_sections))

    for scenario_id in range(no_scenarios):
        for section_id in range(no_sections): # this varies the column of the leontief_mat
            if section_id in EXCLUDE_SECTIONS_INDEX:
                delta_endog = 0
            else:
                delta_endog = (delta[:,scenario_id] * leontief_mat[section_id] / 100 * ALPHA_WEIGHTS/100).sum() # does this still work?
            delta_mat[scenario_id, section_id] = delta_endog

    return delta_mat

def compute_delta(Rt2, delta_mat, delta):
    no_sections = len(SECTIONS)
    no_scenarios = len(Rt2)

    d_mat = np.empty((no_sections, no_scenarios))

    for section_id in range(no_sections):
        for scenario_id in range(no_scenarios):
            if section_id not in EXOGENOUS_SECTIONS_INDEX:
                d = delta_mat[scenario_id, section_id] / (ALPHA_WEIGHTS[section_id]/100)
            else:
                d = delta[section_id, scenario_id]
            d_mat[section_id, scenario_id] = d

    return d_mat

def update_activity(Rt2, d_mat, k_s, r_s):
    no_scenarios = len(Rt2)
    no_sections = len(SECTIONS)

    for section_id in range(no_sections):
        for scenario_id in range(no_scenarios):
            k_val = K_B[section_id] + d_mat[section_id][scenario_id]

            min_k_val = min(1, k_val)
            max_k_val = max(0, min_k_val)
            
            k_s[section_id][scenario_id] = max_k_val

            if section_id not in EXOGENOUS_SECTIONS_INDEX:
                r_val = R_B[section_id]
            else:
                r_val = r_s[section_id][scenario_id]
            max_r_val = max(1, r_val)
            r_s[section_id][scenario_id] = max_r_val

    return k_s, r_s

def simulate_activity(Rt2, workdays, k_s, r_s, duration_s_pred):
    no_sections = len(SECTIONS)
    no_scenarios = len(Rt2)

    k_activity = np.empty((no_sections, no_scenarios, SIMULATION_HORIZON))
    k_workday_activity = np.empty((no_sections, no_scenarios, SIMULATION_HORIZON))

    S_index = workdays[workdays['date'] == KEY_DATES[1]].index[0]

    shutdown_b = workdays[(workdays['date'] >= KEY_DATES[0]) & (workdays['date'] <= KEY_DATES[1])].index
    shutdown_s_endog = workdays[(workdays['date'] > KEY_DATES[1]) & (workdays['date'] <= pd.Timestamp(pd.to_datetime(KEY_DATES[1]) + timedelta(days=int(DURATIONS_S_LEONTIEF))).strftime('%Y-%m-%d'))].index

    for section_id in range(no_sections):
        for scenario_id in range(no_scenarios):
            shutdown_s = workdays[(workdays['date'] > KEY_DATES[1]) & (workdays['date'] <= pd.Timestamp(pd.to_datetime(KEY_DATES[1]) + timedelta(days=int(duration_s_pred[scenario_id]))).strftime('%Y-%m-%d'))].index

            if section_id not in EXOGENOUS_SECTIONS_INDEX:
                shutdown_s = shutdown_s_endog
                duration_s = DURATIONS_S_LEONTIEF
            else:
                shutdown_s = shutdown_s
                duration_s = int(duration_s_pred[scenario_id])

            r_s_m = int(r_s[section_id][scenario_id])
            r_s_w = int(np.around(4*(r_s[section_id][scenario_id] - r_s_m), 0))

            R = S_index + duration_s
            R_val = pd.Timestamp((pd.to_datetime(workdays[workdays.index==S_index].iloc[:,0]) + timedelta(days=int(duration_s))).values[0])

            R_r_val = R_val + relativedelta(months=+r_s_m) + relativedelta(weeks=+r_s_w)
            R_r = workdays[workdays['date'] == str(R_r_val)[:10]].index.values[0]

            case_a_len = len(workdays.loc[(~workdays.index.isin(shutdown_b)) & (~workdays.index.isin(shutdown_s)) & (workdays['date'] <= KEY_DATES[0])])
            case_a = np.ones(case_a_len)

            case_b_len = len(workdays.loc[(workdays.index.isin(shutdown_b))])
            case_b = np.repeat(K_B[section_id], case_b_len)

            case_c_len = len(workdays.loc[(workdays.index.isin(shutdown_s)) & (workdays.index <= workdays.loc[(workdays['date'] == KEY_DATES[1])].index.values[0] + 21)])
            case_c = np.repeat(K_B[section_id], case_c_len) + np.arange(1, case_c_len + 1, 1, dtype=int) * (k_s[section_id][scenario_id] - K_B[section_id]) / (NPI_IMPLEMENTATION_DURATION + 1)

            case_d_len = len(workdays.loc[(workdays.index.isin(shutdown_s)) & (workdays.index > workdays.loc[workdays['date'] == KEY_DATES[1]].index.values[0]+21)])
            case_d = np.repeat(k_s[section_id][scenario_id], case_d_len)

            case_e_len = len(workdays.loc[(~workdays.index.isin(shutdown_b)) & (~workdays.index.isin(shutdown_s)) & (workdays.index > R) & (workdays.index <= R_r)])
            case_e = np.repeat(k_s[section_id][scenario_id], case_e_len) + np.arange(1, case_e_len +1) * ((1-k_s[section_id][scenario_id]) / (R_r - R + 1))

            case_f_len = len(workdays.loc[(~workdays.index.isin(shutdown_b)) & (~workdays.index.isin(shutdown_s)) & (workdays.index > R_r)])
            case_f = np.ones(case_f_len)

            k_activity[section_id, scenario_id, :len(case_a)] = case_a
            k_activity[section_id, scenario_id, len(case_a):len(case_a) + len(case_b)] = case_b
            k_activity[section_id, scenario_id, len(case_a) + len(case_b):len(case_a) + len(case_b) + len(case_c)] = case_c
            k_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c):len(case_a) + len(case_b) + len(case_c) + len(case_d)] = case_d
            k_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c) + len(case_d):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e)] = case_e
            k_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e) + len(case_f)] = case_f

            k_workday_activity[section_id, scenario_id, :len(case_a)] = case_a * workdays['workday'].iloc[:len(case_a)]
            k_workday_activity[section_id, scenario_id, len(case_a):len(case_a) + len(case_b)] = case_b * workdays['workday'].iloc[len(case_a):len(case_a) + len(case_b)]
            k_workday_activity[section_id, scenario_id, len(case_a) + len(case_b):len(case_a) + len(case_b) + len(case_c)] = case_c * workdays['workday'].iloc[len(case_a) + len(case_b):len(case_a) + len(case_b) + len(case_c)]
            k_workday_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c):len(case_a) + len(case_b) + len(case_c) + len(case_d)] = case_d * workdays['workday'].iloc[len(case_a) + len(case_b) + len(case_c):len(case_a) + len(case_b) + len(case_c) + len(case_d)]
            k_workday_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c) + len(case_d):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e)] = case_e * workdays['workday'].iloc[len(case_a) + len(case_b) + len(case_c) + len(case_d):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e)]
            k_workday_activity[section_id, scenario_id, len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e) + len(case_f)] = case_f * workdays['workday'].iloc[len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e):len(case_a) + len(case_b) + len(case_c) + len(case_d) + len(case_e) + len(case_f)]

    return k_activity, k_workday_activity

def agg_costs(workdays, Rt2, k_workday_activity):
    agg_year_month = np.empty((len(workdays['datetime'].dt.year.unique()), len(workdays['datetime'].dt.month.unique()), len(Rt2)))
    
    for year_id, year in enumerate(workdays['datetime'].dt.year.unique()):
        for month_id, month in enumerate(workdays['datetime'].dt.month.unique()):
            start_index = workdays[(workdays['datetime'].dt.year == year) & (workdays['datetime'].dt.month == month)].index.min()
            end_index = workdays[(workdays['datetime'].dt.year == year) & (workdays['datetime'].dt.month == month)].index.max()

            workdays_sum = workdays[(workdays['datetime'].dt.year == year) & (workdays['datetime'].dt.month == month)]['workday'].sum()

            for scenario_id in range(len(Rt2)):
                average_activity_agg = 0
                for section in SECTIONS:
                    k_workday_activity_sum = k_workday_activity[np.where(SECTIONS == section), scenario_id, start_index:end_index+1].sum()
                    average_activity = (k_workday_activity_sum / workdays_sum) * ALPHA_WEIGHTS[np.where(SECTIONS == section)]/100

                    average_activity_agg += average_activity

                agg_year_month[year_id, month_id, scenario_id] = np.round(1-average_activity_agg, decimals=4)
                
    return agg_year_month

def total_costs(costs_sum):
    return costs_sum.sum(axis=(0,1))

def add_country_scenarios(country_dict, Rt2, rel_costs, filename):
    try:
        df = pd.read_csv(filename)
        country_names = df['name']
        R = df['R']
        for i, c in enumerate(country_names):
            idx = np.abs(Rt2 - R[i]).argmin()
            country_dict[c] = (Rt2[idx], rel_costs[idx])
        r = country_dict
    except FileNotFoundError:
        print('No scenario data found')
        r = None
    return r

def plot_rel_costs(
    R,
    costs,
    R_t2=None,
    country_transfer=None,
    colors=['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9',],
    filename=None,
    show=False,
):
    _, ax = pt.subplots(1, 1)
    ax.axhline(y = 0, color = 'C3')
    ax.plot(R, costs)
    if R_t2 is not None:
        ax.set_title(rf'Scenario $R_{{t2}}={R_t2:.2}$')
    if country_transfer is not None:
        for i, c in enumerate(country_transfer):
            ax.scatter(
                country_transfer[c][0],
                country_transfer[c][1],
                color=colors[i],
                label=c,
                zorder=100,
            )
    ax.set_xlabel(r'Reproduction Number $R_{t2}$')
    ax.set_ylabel('Relative Costs')
    ax.set_xlim(0.5, 1)
    ax.set_ylim(None, 70)
    if country_transfer is not None:
        ax.legend()
    pt.tight_layout()
    if filename is not None:
        pt.savefig(filename, dpi=300)
    if show:
        pt.show()
    return ax


if __name__ == '__main__':
    filename_cfg = tools.cmd_line_name()
    with open(filename_cfg) as f:
        cfg = json.load(f)
    dir = cfg['country2']['name']
    filename_epid_scen = Path(f'{dir}/Germany-econ-scenarios.csv')
    filename_epid_refs = Path(f'{dir}/Germany-econ-refs.csv')
    filename_country_scenarios = Path(f'{dir}/Country-scenarios.csv')
    filename_workdays = Path('econ-input/workdays.csv')
    filename_leontief = Path('econ-input/leontief_matrix.csv')

    sir = pd.read_csv(filename_epid_scen)
    Rt2_scen = np.array(sir['Rt2'])
    duration_s_pred_scen = sir['duration']
    Rt2 = np.linspace(0.1, 1.0, 100)
    duration_s_pred = np.interp(
        Rt2,
        Rt2_scen,
        duration_s_pred_scen,
    )
    epid_refs = pd.read_csv(filename_epid_refs)
    r_d_m = np.array(epid_refs['r_d_m'])
    # Rt2 and Rt1 values from reference run, in decreasing order
    r_d_r = np.array(epid_refs['r_d_r'])
    baseline_idx = np.abs(Rt2 - r_d_r[1]).argmin()
    prediction_idx = np.abs(Rt2 - r_d_r[0]).argmin()

    workdays = pd.read_csv(filename_workdays, encoding='utf-8')
    workdays['datetime'] = pd.to_datetime(workdays['date'])

    leontief_matrix = pd.read_csv(filename_leontief, encoding='utf-8')
    leontief_matrix.rename(columns={'Unnamed: 0':'section'}, inplace=True)
    leontief_mat = np.array((leontief_matrix.iloc[:,1:])).T # index selects column

    k_s, r_s, delta = cross_sectional_parameters(Rt2, r_d_m, r_d_r)

    delta_mat = compute_delta_exog(Rt2, delta, leontief_mat)

    d_mat = compute_delta(Rt2, delta_mat, delta)

    k_s, r_s = update_activity(Rt2, d_mat, k_s, r_s)

    k_activity, k_workday_activity = simulate_activity(Rt2, workdays, k_s, r_s, duration_s_pred)

    costs_sum = agg_costs(workdays, Rt2, k_workday_activity)

    scenario_costs = total_costs(costs_sum)

    rel_costs = ((scenario_costs / scenario_costs[baseline_idx] - 1)*100)

    countries = {
        'Germany': (
            Rt2[prediction_idx],
            rel_costs[prediction_idx],
        ),
    }
    countries = add_country_scenarios(
        countries,
        Rt2,
        rel_costs,
        filename_country_scenarios,
    )

    plot_rel_costs(
        Rt2,
        rel_costs,
        r_d_r[1],
        countries,
        ['C0', 'C2', 'C1'],
        Path(f'{dir}/rel_costs_shutdown_run_leontief.png'),
        show=True,
    )
