#!/usr/bin/env python3

from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as pt
import json
import ebesir as sir
import tools


def calc_econ_R(dates, target_date, I_in, avg_window=np.timedelta64(6, 'D')):
    """Calcuate avg. R values according to Dorn et al."""
    dates_r, R = tools.calc_R_RKI_t(dates, I_in)
    dates_r_idx = [
        tools.find_nearest_index(target_date, dates_r),
        tools.find_nearest_index(target_date+avg_window, dates_r)
    ]
    Rt = np.mean(R[dates_r_idx[0]:dates_r_idx[1]])
    return Rt

def get_epid_duration(dates, I_in, threshold, npi_lifting_date, vaccination_date):
    """Calculate duration of epid., depending on thresholds"""
    date_r2_idx = tools.find_nearest_index(npi_lifting_date, dates)
    dates_r = dates[date_r2_idx:]
    I_in_r2 = I_in[date_r2_idx:]
    I_in_low_idx = np.where(I_in_r2 <= threshold)
    if I_in_low_idx[0].size == 0:
        duration = vaccination_date - npi_lifting_date
    else:
        duration = dates_r[I_in_low_idx[0][0]] - npi_lifting_date
    return duration.astype(int)

def get_Rt_ref(filename, dates_r):
    """Calculate Rt ref values from sim. results according to Dorn et al."""
    df = pd.read_csv(filename, parse_dates=['date'])
    # dates = np.array(df['date'], dtype=np.datetime64)
    # I_in = np.array(df['I_in'])
    return (
        calc_econ_R(df['date'], dates_r[0], df['I_in']),
        calc_econ_R(df['date'], dates_r[1], df['I_in']),
    )

def create_ref_output(Rt_ref, filename, duration_ref = [-2.0, 0.0]):
    """Write R ref to file"""
    df_ref_out = pd.DataFrame({
        'r_d_m': duration_ref,
        'r_d_r': Rt_ref,
     })
    df_ref_out.to_csv(filename, index=False)


if __name__ == '__main__':
    filename_cfg = tools.cmd_line_name()
    with open(filename_cfg) as f:
        cfg = json.load(f)
    name = cfg['country1']['name']
    name2 = cfg['country2']['name']

    start_date = cfg['start_date']
    dates_r = [np.datetime64(d) for d in tools.list_of_str_to_dates(cfg['dates_r'])]
    npi_dates = tools.list_of_str_to_dates(cfg['country1']['npi_dates'])[:3]
    npi_dates += [dates_r[0]]
    vaccination_date = np.datetime64(cfg['vaccination_date'])
    end_date = vaccination_date
    npi_lifting_date = np.datetime64(npi_dates[2])
    threshold = cfg['case_threshold']
    impl_duration = cfg['impl_duration']

    min_beta = 0.05 # corresponds roughly to R = 0.1
    max_beta_factor = 1.10
    n_scenarios = 600

    path = Path(name2)
    output_name = f'{name}-econ-scenarios.csv'
    output_ref_name = f'{name}-econ-refs.csv'

    try:
        df_param = pd.read_csv(path / f'{name}-param.csv')
    except FileNotFoundError as e:
        print(f'{e}\n`econ_scenarios.py only works without "transfer_risk_aversion"')
    df_param = df_param.drop(df_param[df_param['name'] == 'beta7'].index) #  pyright: ignore
    df_param = df_param.drop(df_param[df_param['name'] == 'beta6'].index) #  pyright: ignore
    param = list(df_param['value']) #  pyright: ignore
    param[4] = param[3]

    dates, I_obs, N, I0 = sir.get_data(name, start_date, end_date)
    t, npi_t = tools.dates_2_days(dates, npi_dates)

    Rt1_ref, Rt2_ref = get_Rt_ref(path / f'{name}-results.csv', dates_r)

    scenario = []
    duration = []
    Rt1 = []
    Rt2 = []
    tmp_param = param[:]
    for i in range(n_scenarios):
        scenario.append(i)
        tmp_param[4] = (
            min_beta + (max_beta_factor-min_beta) * i / n_scenarios * param[4]
        )
        I_in, S, I, tau, beta = sir.solve_besir_model(
            sir.besir,
            t,
            I0,
            N,
            npi_t,
            tmp_param,
            impl_duration,
        )

        duration.append(get_epid_duration(dates, I_in, threshold, npi_lifting_date, vaccination_date))

        Rt1.append(calc_econ_R(dates, dates_r[0], I_in))
        Rt2.append(calc_econ_R(dates, dates_r[1], I_in))

        print(f'Scenario {i+1}, contact rate \u03B2 = {tmp_param[-3]:.3}, reference: {param[-3]:.3}')
        print(f'Duration: {duration[i]} days')
        try:
            print(f'R-value: Rt1 = {Rt1[i]:.3}, Rt2 = {Rt2[i]:.3}, reference: {Rt2_ref:.3}')
        except NameError:
            print(f'R-value: Rt1 = {Rt1[i]:.3}, Rt2 = {Rt2[i]:.3}')
        print('')
    Rt1 = np.asarray(Rt1)
    Rt2 = np.asarray(Rt2)

    duration = np.asarray(duration)

    # Dorn et al. R values for scenarios
    # match them for writing duration vs R for econ model
    R_output = np.asarray((
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.528,
        0.6,
        0.6587,
        0.7005,
        0.75,
        0.8,
        0.81,
        0.82,
        0.83,
        0.84,
        0.8496,
        0.85,
        0.86,
        0.87,
        0.88,
        0.89,
        0.9,
        0.91,
        0.92,
        0.93,
        0.94,
        0.95,
        1.0,
    ))
    Rt2_output = np.empty_like(R_output)
    duration_output = np.empty_like(R_output)
    for i, r in enumerate(R_output):
        idx = (np.abs(Rt2 - r)).argmin()
        Rt2_output[i] = Rt2[idx]
        duration_output[i] = duration[idx]

    df_output = pd.DataFrame({
        'Scenario': range(len(R_output)),
        'Rt1': np.tile(Rt1_ref, len(R_output)),
        'Rt2': Rt2_output,
        'duration': duration_output,
    })

    create_ref_output([Rt2_ref, Rt1_ref], path / output_ref_name)
    df_output.to_csv(path / output_name, index=False)
