from datetime import timedelta
import matplotlib.pyplot as pt
from matplotlib import rcParams
import seaborn as sns


rcParams["figure.autolayout"] = True
LW = 3.0
F_SIZE = 12
FT_SIZE = 11
LW_VERT_LINES = 1.0
ALPHA_VERT_LINES = 0.5


def npis(npi_dates=None, param_trans=None, ax=None):
    if ax is None:
        ax = pt
    linestyle = ['-'] * len(npi_dates)
    # use dashed lines for periods where no params where transfered
    if param_trans is not None and any(param_trans):
        linestyle_choice = ['--', '-']
        linestyle = [linestyle_choice[b] for b in param_trans]

    if npi_dates is not None:
        [ax.axvline(
            t-timedelta(days=1),
            color='k',
            linestyle=linestyle[i],
            linewidth=LW_VERT_LINES,
            alpha=ALPHA_VERT_LINES,
            zorder=-100
        ) for i, t in enumerate(npi_dates)]

def _check_ax(ax):
    if ax is None:
        fig, ax = pt.subplots(1, 1)
    else:
        fig = ax.get_figure()
    return fig, ax

def _title_and_save(ax, title=None, savename=None):
    if title is not None:
        ax.set_title(title)
    if savename is not None:
        pt.savefig(savename, dpi=300)

def cases(
    dates,
    I_in,
    I_obs,
    npi_dates=None,
    param_trans=None,
    title=None,
    ax=None,
    savename=None,
    incidences=False,
):
    fig, ax = _check_ax(ax)
    ax.scatter(dates, I_obs, label='Observations')
    ax.plot(dates, I_in, label='Model')
    ax.set_xlabel('Dates')
    if incidences:
        ax.set_ylabel('7 day Incidence')
        # manually set ylim to align risk tranfser with non risk transfer
        # ax.set_ylim(0, 51)
    else:
        ax.set_ylabel('Reported Cases per Day')
    fig.autofmt_xdate()
    npis(npi_dates, param_trans, ax)
    _title_and_save(ax, title, savename)
    return ax

def R(
    dates,
    R,
    R_obs,
    R_param=None,
    npi_dates=None,
    param_trans=None,
    title=None,
    ax=None,
    savename=None,
    df_external=None,
    external_label=None,
    highlight_dates=None,
):
    c = sns.color_palette()
    fig, ax = _check_ax(ax)
    ax.plot(dates, R_obs, color=c[0], label='From Reported Cases')
    ax.plot(dates, R, color=c[0], linestyle='--', label='From Modelled Cases')
    if R_param is not None:
        ax.plot(dates, R_param, color=c[1], linestyle='--', label='From Model Parameter')
    if df_external is not None:
        pt.plot(
            df_external['date'],
            df_external['R'],
            color=c[2],
            label=external_label,
        )
        try:
            pt.fill_between(
                df_external['date'],
                df_external['R_lower'],
                df_external['R_upper'],
                color=c[2],
                alpha=0.3,
            )
        except KeyError:
            pass
    ax.set_ylim(0)
    ax.set_xlabel('Dates')
    ax.set_ylabel('Reproduction Number $R$')
    fig.autofmt_xdate()
    npis(npi_dates, param_trans, ax)
    if highlight_dates is not None:
        import matplotlib.patches as patches
        d = highlight_dates[0]
        rect = patches.Rectangle((d[0], 0.7), d[1]-d[0], 0.3, alpha=0.4)
        ax.add_patch(rect)
        d = highlight_dates[1]
        rect = patches.Rectangle((d[0], 0.75), d[1]-d[0], 0.45, alpha=0.4)
        ax.add_patch(rect)
    ax.legend()
    _title_and_save(ax, title, savename)
    return ax
