#!/usr/bin/env python

import argparse
from datetime import datetime
import numpy as np
import pandas as pd
import pycocap as pyc


pd.options.mode.chained_assignment = None  # default='warn'


def find_nearest_value(value, array):
    """Find nearest match of given value in given array.

    Args:
        value: number to find in array
        array: search for nearest match in this numpy array
    """
    return min(array, key = lambda x: abs(x - value))

def find_nearest_index(value, array):
    """Find index of nearest match of given value in given array.

    Args:
        value: index of this number will be found in array
        array: search for nearest match in this numpy array
    """
    return list(array).index(find_nearest_value(value, array))

def str_to_date(string, format='%Y-%m-%d'):
    """Convert a string in given format to Python date"""
    return datetime.strptime(string, format).date()

def list_of_str_to_dates(lst, format='%Y-%m-%d'):
    """Convert a list of strings in given format to Python dates"""
    return [str_to_date(d, format) for d in lst]

def cmd_line_name():
    parser = argparse.ArgumentParser(description='A coupled behavioral - SIR model')
    parser.add_argument(
        'Country_filename', help='Specify the Country JSON filename with which to calibrate the model.'
    )
    name = parser.parse_args().Country_filename

    return name

def write_param(param, param_names, filename):
    """Write parameters to file"""
    df = pd.DataFrame({'name': param_names, 'value': param})
    df.to_csv(filename, index=False)

def write_data(
    name,
    dates,
    npi_dates,
    start_date,
    end_date,
    N,
    impl_duration,
    I_in,
    S,
    I,
    I_obs,
    tau,
    beta,
    gamma,
    v,
    filename,
    transfer_param_ind=None,
):
    """Write all relevant output data to file"""
    inc = calc_incidence(I_in, N)
    inc_obs = calc_incidence(I_obs, N)

    df = pyc.npi(
        name,
        'stringency',
        start_date,
        end_date,
    )

    stringency = np.asarray(df['stringency'])

    param_trans = np.full(I.shape, False)

    if transfer_param_ind is not None:
        idx1 = find_nearest_index(np.datetime64(npi_dates[transfer_param_ind[0]-1]), dates)
        idx2 = None
        if transfer_param_ind[1] <= len(npi_dates):
            idx2 = find_nearest_index(np.datetime64(npi_dates[transfer_param_ind[1]-2]), dates)

        param_trans[idx1:idx2] = True

    df = pd.DataFrame({
        'date': dates,
        'I_in': I_in,
        'S': S,
        'I': I,
        'I_obs': I_obs,
        'incidence': inc,
        'incidence_obs': inc_obs,
        'stringency': stringency,
        'tau': tau,
        'beta': beta,
        'gamma': gamma,
        'v': v,
        'impl_duration': impl_duration,
        'N': N,
        'parameter_transfer': param_trans,
    })

    df.to_csv(filename, index=False)
    return df

def write_npi_dates(npi_dates, filename, param_trans_idx=None):
    """Write NPI dates to file"""
    df = pd.DataFrame({'npi_dates': npi_dates})
    df['parameter_transfer'] = False
    if param_trans_idx is not None:
        df['parameter_transfer'][param_trans_idx[0]-1:param_trans_idx[1]] = True
    df.to_csv(filename, index=False)
    return df

def dates_2_days(dates, npi_dates=None):
    """Convert dates to integer days, converts NPI dates too, if given"""
    dates_np = [np.datetime64(d) for d in dates]
    t = np.array([
        (d - dates_np[0]).astype('timedelta64[D]').astype(int) for d in dates_np
        ])
    r = t
    if npi_dates is not None:
        npi_dates_np = [np.datetime64(d) for d in npi_dates]
        npi_t = [
            (d - dates_np[0]).astype('timedelta64[D]').astype(int)
                for d in npi_dates_np
        ]
        r = t, npi_t
    return r

def calc_incidence(cases, N, size=7):
    """Calculate `size` day incidence rate"""
    return (
        np.array(
            pd.DataFrame(cases, columns=['c'])['c'].
            rolling(window=size).sum() / N * 100_000.
        )
    )

def calc_R_RKI_t(dates, I_obs):
    """Calculate the reproduction value R for the end of a timeseries.

    This method is used by the RKI.
    """
    I_7d = smooth(I_obs, 7, 'flat')
    R0_RKI_7 = []
    for i in range(12, len(I_obs)):
        I_7d_4d_sum = np.sum(I_7d[i-8:i-4])
        I_7d_8d_sum = np.sum(I_7d[i-12:i-8])
        R0_RKI_7.append(I_7d_4d_sum / I_7d_8d_sum if I_7d_8d_sum > 0 else 0.0)

    return dates[12:], np.asarray(R0_RKI_7)

def smoothstep(x, y1, y2, x1):
    """ Create a smooth interpolation between y1 and y2 from x to x+x1 """
    return (
        y1 + (y2 - y1) *
        (lambda xc: np.where(xc < 0 , 0, np.where(xc <= 1, 3 * xc**2 - 2 *xc**3, 1)))
            (x / x1)
    )

def param_piecewise(t, param, npi_t=[], impl_duration=1):
    """Create continous piecew. const. fct.s, with option to smooth jumps"""
    npi_periods = npi_t[:]
    npi_periods.insert(0, t[0])
    npi_periods.append(t[-1])
    p_ext = [param[0]] + param[:]
    p_ext = np.insert(param, 0, param[0])
    p_tiled = []
    for i in range(len(npi_periods)-1):
        p_tiled.extend(
            [
                smoothstep(p, p_ext[i], p_ext[i+1], impl_duration)
                for p in range(int(npi_periods[i+1] - npi_periods[i]))]
        )
    # prevent index problems in final time steps
    p_tiled.append(p_tiled[-1])
    return np.asarray(p_tiled)

def smooth(x, window_len=11, window='hanning'):
    """Smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this:
    return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len<3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    if window == 'flat': #moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('numpy.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')
    if len(y) % 2 == 0:
        # TODO check why this does not work anymore (wrong len)
        # idx0 = window_len // 2 - 1
        idx0 = window_len // 2
    else:
        idx0 = window_len // 2
    return y[idx0:-(window_len//2)]

